#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

import re

from .config import VK
from .utils.another import unixToDatetime


class Artist:
    """
    Класс, представляющий артиста.

    Атрибуты:
        nickname (str): псевдоним артиста.\n
        photo (dict, optional): словарь с размерами и URL фотографий артиста, отсортированный по размеру.\n
        albums (list[Album], optional): список альбомов артиста, представленных объектами класса `Album`.\n
        tracks (list[Track], optional): список аудиотреков артиста, представленных объектами класса `Track`.\n
        id (str): идентификатор артиста.\n
        url (str): URL страницы артиста.
    """

    def __init__(self, artist: dict) -> None:
        self.nickname = artist.get("name")

        photo = artist.get("photo")
        if photo:
            photoDict = {f"{photo.get('width')}": re.sub(r"\.(j|jp|jpg)$", "", photo.get("url")[:photo.get("url").rfind("&c_uniq_tag=")]) for photo in photo}
            self.photo = dict(sorted(photoDict.items(), key=lambda item: (int(item[0]))))

        else:
            photo = artist.get("photos")
            if photo:
                photoDict = {f"{photo.get('width')}": photo.get("url") for photo in photo[0]["photo"]}
                self.photo = dict(sorted(photoDict.items(), key=lambda item: (int(item[0]))))

        albums = artist.get("albums")
        if albums:
            self.albums = [Album(album) for album in albums]

        tracks = artist.get("tracks")
        if tracks:
            self.tracks = [Track(track) for track in tracks]

        if self.nickname != "Various Artists":
            self.id = artist.get("id")
            domain = artist.get("domain")
            self.url = VK + "artist/" + (domain if domain else self.id)


class Album:
    """
    Класс, представляющий альбом.

    Атрибуты:
        title (str): название альбома.\n
        subtitle (str, optional): подзаголовок альбома, если он присутствует.\n
        description (str, optional): описание альбома, если оно присутствует.\n
        artists (list[Artist]): список основных артистов альбома, представленных объектами класса `Artist`.\n
        featuredArtists (list[Artist], optional): список приглашённых артистов альбома, представленных объектами класса `Artist`.\n
        releaseYear (int, optional): год выпуска альбома.\n
        genres (list[Genre], optional): список жанров альбома, представленных объектами класса Genre.\n
        plays (int, optional): количество прослушиваний альбома.\n
        uploadedAt (datetime, optional): дата и время загрузки альбома.\n
        updatedAt (datetime, optional): дата и время последнего обновления информации об альбоме.\n
        photo (dict, optional): словарь с размерами и URL фотографий альбома, отсортированный по размеру.\n
        tracks (list[Track], optional): список аудиотреков альбома, где каждый аудиотрек представлен объектом класса `Track`.\n
        exclusive (bool, optional): флаг, указывающий, является ли альбом эксклюзивным.\n
        ownerId (str): идентификатор владельца альбома.\n
        albumId (str): идентификатор альбома.\n
        id (str): комбинированный идентификатор в формате `ownerId_albumId`.\n
        url (str): URL страницы альбома.
    """

    def __init__(self, album: dict, playlist: bool = False) -> None:
        title = album.get("title")
        if title:
            self.title = title
        subtitle = album.get("subtitle")
        if subtitle:
            self.subtitle = subtitle

        description = album.get("description")
        if description:
            self.description = description

        mainArtists = album.get("main_artists")
        if mainArtists:
            self.artists = [Artist(mainArtist) for mainArtist in mainArtists]

        featuredArtists = album.get("featured_artists")
        if featuredArtists:
            self.featuredArtists = [Artist(featuredArtist) for featuredArtist in featuredArtists]

        releaseYear = album.get("year")
        if releaseYear:
            self.releaseYear = releaseYear

        genres = album.get("genres")
        if genres:
            for index, genre in enumerate(genres):
                genres[index] = Genre(genre)
            self.genres = genres

        plays = album.get("plays")
        if plays:
            self.plays = plays

        uploadedAt = album.get("create_time")
        if uploadedAt:
            self.uploadedAt = unixToDatetime(uploadedAt)

        updatedAt = album.get("update_time")
        if updatedAt:
            self.updatedAt = unixToDatetime(updatedAt)

        photo = album.get("photo")
        if not photo:
            photo = album.get("thumb")

        if photo:
            self.photo = {key.split("_")[1]: value[:value.rfind("&c_uniq_tag=")] for key, value in photo.items() if key.startswith("photo_")}

        original = album.get("original")
        if original:
            self.original = Album(original)

        tracks = album.get("tracks")
        if tracks:
            self.tracks = tracks

        exclusive = album.get("exclusive")
        if exclusive is not None:
            self.exclusive = exclusive

        self.ownerId = album.get("owner_id")
        if not playlist:
            self.albumId = album.get("id") or album.get("playlist_id")
            self.id = f"{self.ownerId}_{self.albumId}"
            self.url = VK + "music/album/" + self.id

        else:
            self.playlistId = album.get("id")
            self.id = f"{self.ownerId}_{self.playlistId}"
            self.url = VK + "music/playlist/" + self.id


class Track:
    """
    Класс, представляющий аудиотрек.

    Атрибуты:
        title (str): название аудиотрека.\n
        subtitle (str, optional): подзаголовок аудиотрека, если он присутствует.\n
        artists (list[Artist], optional): список основных артистов аудиотрека, представленных объектами класса `Artist`.\n
        artist (str): основной(ые) артист(ы) аудиотрека.\n
        featuredArtists (list[Artist], optional): список приглашённых артистов аудиотрека, представленных объектами класса `Artist`.\n
        genre (Genre, optional): жанр аудиотрека, представленный объектом класса `Genre`.\n
        explicit (bool, optional): флаг, указывающий, есть ли в треке ненормативная лексика.\n
        duration (int): продолжительность аудиотрека в секундах.\n
        fileUrl (str, optional): ссылка на MP3-файл.\n
        lyrics (str, optional): текст аудиотрека.\n
        hasLyrics (bool, optional): флаг, указывающий, имеет ли аудиотрек текст.\n
        uploadedAt (datetime, optional): дата и время загрузки аудиотрека.\n
        album (Album, optional): альбом, к которому принадлежит аудиотрек, представленный объектом класса `Album`.\n
        ownerId (str): идентификатор владельца трека.\n
        trackId (str): идентификатор аудиотрека.\n
        id (str): комбинированный идентификатор в формате `ownerId_trackId`.\n
        url (str): URL страницы аудиотрека.
    """

    def __init__(self, track: dict) -> None:
        self.title = track.get("title")
        subtitle = track.get("subtitle")
        if subtitle:
            self.subtitle = subtitle

        mainArtists = track.get("main_artists")
        if mainArtists:
            self.artists = [Artist(mainArtist) for mainArtist in mainArtists]

        artist = track.get("artist")
        if artist:
            self.artist = artist

        featuredArtists = track.get("featured_artists")
        if featuredArtists:
            self.featuredArtists = [Artist(featuredArtist) for featuredArtist in featuredArtists]

        genreId = track.get("genre_id")
        if genreId:
            self.genre = Genre(genreId=genreId)

        explicit = track.get("is_excplicit")
        if explicit is not None:
            self.explicit = explicit

        self.duration = track.get("duration")

        fileUrl = track.get("url")
        if fileUrl:
            self.fileUrl = fileUrl[:fileUrl.rfind("?siren=1")]

        lyrics = track.get("lyrics")
        if lyrics:
            self.lyrics = lyrics

        else:
            hasLyrics = track.get("has_lyrics")
            if hasLyrics is not None:
                self.hasLyrics = hasLyrics

        uploadedAt = track.get("date")
        if uploadedAt:
            self.uploadedAt = unixToDatetime(uploadedAt)

        album = track.get("album")
        if album:
            self.album = Album(album)

        licensed = track.get("is_licensed")
        if licensed is not None:
            self.licensed = licensed

        self.ownerId = track.get("owner_id")
        self.trackId = track.get("id")
        self.id = f"{self.ownerId}_{self.trackId}"

        releaseAudioId = track.get("release_audio_id")
        self.url = VK + "audio" + (self.id if not releaseAudioId else releaseAudioId)


class Playlist:
    """
    Класс, представляющий плейлист.

    Атрибуты:
        title (str): название плейлиста.\n
        subtitle (str, optional): подзаголовок плейлиста, если он присутствует.\n
        description (str, optional): описание плейлиста, если оно присутствует.\n
        plays (int, optional): количество прослушиваний плейлиста.\n
        createdAt (datetime, optional): дата и время создания плейлиста.\n
        updatedAt (datetime, optional): дата и время последнего добавления аудиотрека в плейлист (или удаления из него).\n
        photo (dict, optional): словарь с размерами и URL фотографий плейлиста, отсортированный по размеру.\n
        tracks (list[Track], optional): список аудиотреков плейлиста, где каждый аудиотрек представлен объектом класса `Track`.\n
        ownerId (str): идентификатор владельца плейлиста.\n
        playlistId (str): идентификатор плейлиста.\n
        id (str): комбинированный идентификатор в формате `ownerId_playlistId`.\n
        url (str): URL страницы плейлиста.
    """

    def __init__(self, playlist: dict, isOwn: bool = False) -> None:
        title = playlist.get("title")
        if title:
            self.title = title
        subtitle = playlist.get("subtitle")
        if subtitle:
            self.subtitle = subtitle

        description = playlist.get("description")
        if description:
            self.description = description

        plays = playlist.get("plays")
        if plays:
            self.plays = plays

        createddAt = playlist.get("create_time")
        if createddAt:
            self.createdAt = unixToDatetime(createddAt)

        updatedAt = playlist.get("update_time")
        if updatedAt:
            self.updatedAt = unixToDatetime(updatedAt)

        photo = playlist.get("photo")
        if not photo:
            photo = playlist.get("thumb")

        if photo:
            self.photo = {key.split("_")[1]: value[:value.rfind("&c_uniq_tag=")] for key, value in photo.items() if key.startswith("photo_")}

        original = playlist.get("original")
        if original:
            self.original = Playlist(original)

        tracks = playlist.get("tracks")
        if tracks:
            self.tracks = tracks

        self.ownerId = playlist.get("owner_id")
        self.playlistId = playlist.get("id") or playlist.get("playlist_id")
        self.id = f"{self.ownerId}_{self.playlistId}"
        self.url = VK + "music/playlist/" + self.id

        self.own = isOwn


trackGenres = {
    1: "Рок",
    2: "Поп",
    3: "Рэп и Хип-хоп",
    4: "Расслабляющая",
    5: "House и Танцевальная",
    6: "Инструментальная",
    7: "Метал",
    8: "Дабстеп",
    10: "Drum & Bass",
    11: "Транс",
    12: "Шансон",
    13: "Этническая",
    14: "Акустическая",
    15: "Регги",
    16: "Классическая",
    17: "Инди-поп",
    18: "Другая",
    19: "Скит",
    21: "Альтернатива",
    22: "Электро-поп и Диско",
    1001: "Джаз и Блюз"
}


class Genre:
    """
    Класс, представляющий жанр аудиотрека или альбома.

    Атрибуты:
        title (str): название жанра.\n
        id (str): идентификатор жанра.\n
    """

    def __init__(self, genre: dict = None, genreId: int = None) -> None:
        if genreId:
            self.title = trackGenres.get(genreId, "Неизвестен")

            self.id = genreId

        else:
            self.title = genre.get("name")

            self.id = genre.get("id")



#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

import os

import pytz
from datetime import datetime

from typing import Union, Optional, List, Tuple

from ..models import Track, Playlist
from ..config import moscowTz
from ..utils.another import checkFile


class Update:
    def upload(self, filename: str, title: str = None, artist: str = None, lyrics: str = None, genreId: int = None, removeFromSearchResults: bool = None, groupId: int = None) -> Union[Track, dict]:
        """
        Загружает новый аудиотрек во ВКонтакте.

        Пример загрузки файла с названием «prombl — zapreti.mp3»:\n
        result = client.upload(filename="prombl — zapreti", title="zapreti", artist="prombl", lyrics="yourLyrics", removeFromSearchResults=True, groupId="yourGroupId")\n
        print(result)

        :param filename: имя MP3-файла, содержащего аудиотрек, который необходимо загрузить (без расширения). (str)
        :param title: название аудиотрека. (str, необязательно)
        :param artist: артист(ы) аудиотрека. (str, необязательно)
        :param lyrics: текст аудиотрека. (str, необязательно, временно не работает)
        :param genreId: жанр аудиотрека (в виде идентификатора). (int, необязательно, временно не работает)
        :param removeFromSearchResults: флаг, указывающий, будет ли аудиотрек скрыт из поисковой выдачи. (bool, необязательно, временно не работает)
        :param groupId: идентификатор группы, в музыку которой необходимо загрузить аудиотрек (int, необязательно, временно не работает)
        :return: загруженный аудиотрек в виде объекта модели `Track` или словаря.
        """

        if filename.endswith(".mp3"):
            filename = filename[:-4]

        filename = checkFile(filename + ".mp3")
        if not filename:
            self._raiseError("MP3FileNotFound")

        fileSizeInBytes = os.path.getsize(filename)
        fileSizeInMB = fileSizeInBytes / (1024 * 1024)  # Преобразование размера файла в мегабайты

        if fileSizeInMB > 200:  # Проверка, меньше ли размер файла 200 MB
            self._raiseError("MP3FileTooBig")

        uploadUrl = self._VKReq("getUploadServer").get("upload_url")

        with open(filename, "rb") as file:
            fileContent = file.read()

        uploadingFileResponse = self._client.sendReq(uploadUrl, files={"file": (filename, fileContent, "audio/mpeg")}, method="POST")

        server = uploadingFileResponse.get("server")
        audio = uploadingFileResponse.get("audio")
        hash = uploadingFileResponse.get("hash")

        response = self._VKReq("save", {"server": server, "audio": audio, "hash": hash, "title": title, "artist": artist})

        return self._finalizeResponse(response, Track) if not response.get("error") else response


    def add(self, ownerIds: Union[int, List[int]], trackIds: Union[int, List[int]], playlistId: Optional[int] = None, groupId: Optional[int] = None) -> Tuple[Union[bool, dict]]:
        """
        Добавляет аудиотрек в музыку или плейлист пользователя или группы.

        Пример использования:\n
        result = client.add(ownerIds=474499244, trackIds=456638035, playlistId="yourPlaylistId", groupId="yourGroupId")\n
        print(result)

        :param ownerIds: идентификатор(ы) владельца аудиотрека(ов) (пользователь или группа). (int или list)
        :param trackIds: идентификатор(ы) аудиотрека(ов), который(е) необходимо добавить. (int или list)
        :param playlistId: идентификатор плейлиста, в который необходимо добавить аудиотрек. (int, необязательно)
        :param groupId: идентификатор группы, в музыку или плейлист которой необходимо добавить аудиотрек. (int, необязательно)
        :return: `True`, если аудиотрек успешно добавлен, `False` в противном случае.
        """

        if type(ownerIds) != type(trackIds):
            return self._raiseError("ownerIdsAndTrackIdsTypeDifferent")

        if isinstance(ownerIds, list) and isinstance(trackIds, list) and len(ownerIds) != len(trackIds):
            return self._raiseError("ownerIdsAndTrackIdsLenDifferent")

        if not (isinstance(ownerIds, list) and isinstance(trackIds, list)):
            ownerIds = [ownerIds]
            trackIds = [trackIds]

        if not groupId:
            groupId = self.getSelf().get("id")

        if playlistId:
            method = "addToPlaylist"

        else:
            method = "add"

        results = []
        for ownerId, trackId in zip(ownerIds, trackIds):
            response = self._VKReq(method, {**({"owner_id": ownerId, "audio_id": trackId, "group_id": groupId} if not playlistId else {"owner_id": groupId, "audio_ids": f"{ownerId}_{trackId}"}), "playlist_id": playlistId})
            results.append(bool(response) if not (isinstance(response, dict) and response.get("error")) else self._raiseError("playlistNotFound"))

        return tuple(results)


    def remove(self, ownerIds: Union[int, List[int]], trackIds: Union[int, List[int]], playlistId: Optional[int] = None, groupId: Optional[int] = None, reValidateIds: bool = True) -> Tuple[Union[bool, dict]]:
        """
        Удаляет аудиотрек из музыки или плейлиста пользователя или группы.

        Пример использования:\n
        result = client.remove(ownerIds=474499244, trackIds=456638035, playlistId="yourPlaylistId", groupId="yourGroupId", reValidateIds=False)\n
        print(result)

        :param ownerIds: идентификатор(ы) владельца аудиотрека(ов) (пользователь или группа). (int или list)
        :param trackIds: идентификатор(ы) аудиотрека(ов), который(е) необходимо удалить. (int или list)
        :param playlistId: идентификатор плейлиста, из которого необходимо удалить аудиотрек(и). (int, необязательно, метод временно не работает для плейлистов, привязанных к чату)
        :param groupId: идентификатор группы, из музыки или плейлиста которой необходимо удалить аудиотрек(и). (int, необязательно)
        :param reValidateIds: флаг, указывающий, необходимо ли перепроверить идентификатор(ы) аудитрека(ов) по находящимся в плейлисте. (bool, по умолчанию `True`)
        :return: `True`, если аудиотрек(и) успешно удалён, `False` в противном случае.
        """

        if type(ownerIds) != type(trackIds):
            return self._raiseError("ownerIdsAndTrackIdsTypeDifferent")

        if isinstance(ownerIds, list) and isinstance(trackIds, list) and len(ownerIds) != len(trackIds):
            return self._raiseError("ownerIdsAndTrackIdsLenDifferent")

        if not (isinstance(ownerIds, list) and isinstance(trackIds, list)):
            ownerIds = [ownerIds]
            trackIds = [trackIds]

        if not groupId:
            groupId = self.getSelf().get("id")

        if reValidateIds:
            if playlistId:
                playlist = self.getPlaylist(playlistId, groupId, includeTracks=True)
                if isinstance(playlist, dict) and playlist.get("error"):
                    return playlist

                if isinstance(playlist, dict):
                    existTracks = playlist.get("tracks")

                elif isinstance(playlist, Playlist):
                    existTracks = playlist.tracks if hasattr(playlist, "tracks") else None

            else:
                existTracks = self.getTracks(groupId)

            if not existTracks and playlistId:
                return (False,)

            for index, (ownerId, trackId) in enumerate(zip(ownerIds, trackIds)):
                if not playlistId:
                    continue

                track = self.get(ownerId, trackId)

                if isinstance(track, dict):
                    if track.get("error"):
                        continue

                    trackTitle = track.get("title")
                    trackArtists = track.get("artists")
                    if not trackArtists:
                        trackArtists = track.get("artist")

                elif isinstance(track, Track):
                    trackTitle = track.title
                    trackArtists = track.artists
                    if not trackArtists:
                        trackArtists = track.artist

                for existTrack in existTracks:
                    if isinstance(existTrack, dict):
                        existTrackTitle = existTrack.get("title")
                        existTrackArtists = existTrack.get("artists")
                        if not existTrackArtists:
                            existTrackArtists = existTrack.get("artist")

                        existTrackOwnerId = existTrack.get("ownerId")
                        existTrackId = existTrack.get("trackId")

                    elif isinstance(existTrack, Track):
                        existTrackTitle = existTrack.title
                        existTrackArtists = existTrack.artists
                        if not existTrackArtists:
                            existTrackArtists = existTrack.artist

                        existTrackOwnerId = existTrack.ownerId
                        existTrackId = existTrack.trackId

                    if existTrackTitle == trackTitle and existTrackArtists == trackArtists:
                        ownerIds[index] = existTrackOwnerId
                        trackIds[index] = existTrackId
                        break

        if playlistId:
            def removeTrackFromPlaylist(id: str) -> bool:
                response = self._VKReq("removeFromPlaylist", {"audio_ids": id, "owner_id": groupId if groupId > 0 else groupId - groupId * 2, "playlist_id": playlistId})
                return bool(response) if not (isinstance(response, dict) and response.get("error")) else response

            tasks = [removeTrackFromPlaylist(f"{ownerId}_{trackId}") for ownerId, trackId in zip(ownerIds, trackIds)]

        else:
            def removeTrack(ownerId: int, trackId: int) -> bool:
                response = self._VKReq("delete", {"owner_id": ownerId, "audio_id": trackId, "group_id": groupId})
                return bool(response) if not (isinstance(response, dict) and response.get("error")) else response

            tasks = [removeTrack(ownerId, trackId) for ownerId, trackId in zip(ownerIds, trackIds)]

        return tuple(tasks)


    def edit(self, ownerId: int, trackId: int, title: Union[str, int] = None, artist: Union[str, int] = None, lyrics: Union[str, int] = None, genreId: int = None, removeFromSearchResults: bool = None) -> Union[bool, dict]:
        """
        Изменяет информацию об аудиотреке.

        Пример использования:\n
        result = client.edit(ownerId="yourOwnerId", trackId="yourTrackId", title="zapreti", artist="prombl", "lyrics"="", "genreId"=3, removeFromSearchResults=True)\n
        print(result)

        :param ownerId: идентификатор владельца аудиотрека (пользователь или группа). (int)
        :param trackId: идентификатор аудиотрека, информацию которого необходимо изменить. (int)
        :param title: новое название аудиотрека. (str, необязательно)
        :param artist: новый(е) артист(ы) аудиотрека. (str, необязательно)
        :param lyrics: новый текст аудиотрека. (str, необязательно)
        :param genreId: новый жанр аудиотрека (в виде идентификатора). (int, необязательно)
        :param removeFromSearchResults: флаг, указывающий, будет ли аудиотрек скрыт из поисковой выдачи. (bool, необязательно)
        :return: `True`, если информация аудиотрека успешно обновлена, `False` в противном случае.
        """

        if not any((title, artist, lyrics is not None, genreId is not None, removeFromSearchResults is not None)):
            return False

        response = self._VKReq("edit", {"owner_id": ownerId, "audio_id": trackId,
            **({"title": title} if title else {}),
            **({"artist": artist} if artist else {}),
            **({"lyrics": lyrics} if lyrics is not None else {}),
            **({"genre_id": genreId} if genreId is not None else {}),
            **({"no_search": removeFromSearchResults} if removeFromSearchResults is not None else {})
        })

        return bool(response) if not (isinstance(response, dict) and response.get("error")) else response


    def restore(self, ownerId: int, trackId: int) -> Union[bool, dict]:
        """
        Восстанавливает удалённый аудиотрек.

        Пример использования:\n
        result = client.restore(ownerId="yourOwnerId", trackId="yourTrackId")\n
        print(result)

        :param ownerId: идентификатор владельца аудиотрека (пользователь или группа). (int)
        :param trackId: идентификатор аудиотрека, который необходимо восстановить. (int)
        :return: `True`, если аудиотрек успешно восстановлен, `False` в противном случае.
        """

        response = self._VKReq("restore", {"owner_id": ownerId, "audio_id": trackId})

        return bool(response) if not (isinstance(response, dict) and response.get("error")) else response


    def createPlaylist(self, title: Union[str, int], description: Union[str, int] = None, photo: str = None, groupId: int = None, chatId: int = None) -> Union[int, dict, None]:
        """
        Создаёт плейлист в музыке пользователя или группы.

        Пример использования:\n
        result = client.createPlaylist(title="prombl — npc", description="Release Date: December 24, 2021", "photo"="yourPhotoUrl", "groupId"="yourGroupId", chatId="yourChatId")\n
        print(result)

        :param title: название плейлиста. (str)
        :param description: описание плейлиста. (str, необязательно)
        :param photo: фото плейлиста. (str, необязательно, временно не работает)
        :param groupId: идентификатор группы, в которой необходимо создать плейлист. (int, необязательно)
        :param chatId: идентификатор чата, к которому привязать плейлист. (int, формат: `2000000000 + идентификатор чата`)
        :return: идентификатор плейлиста или словарь с ключами `ownerId` и `playlistId` (если создаётся плейлист, привязанный к чату), если плейлист успешно создан, `None` в противном случае.
        """

        if not groupId:
            groupId = self.getSelf().get("id")

        if chatId:
            method = "createChatPlaylist"

        else:
            method = "createPlaylist"

        playlist = self._VKReq(method, {"title": title, "description": description, "owner_id": groupId})
        playlistId = playlist.get("id")

        if playlistId:
            if chatId:
                playlistOwnerId = playlist.get("owner_id")
                return {"ownerId": playlistOwnerId, "playlistId": playlistId}

            return playlistId

            """if photo:
                self._editPlaylistPhoto(playlistId, photo, groupId)"""

        return playlist


    def addPlaylist(self, playlistId: int, ownerId: int = None, groupId: int = None) -> Union[int, None, dict]:
        """
        Добавляет плейлист в музыку пользователя или группы.

        Пример использования:\n
        result = client.addPlaylist(playlistId=1, ownerId=-215973356, groupId="yourGroupId")\n
        print(result)

        :param ownerId: идентификатор владельца плейлиста (пользователь или группа). (int, по умолчанию текущий пользователь)
        :param playlistId: идентификатор плейлиста, который необходимо добавить. (int)
        :param groupId: идентификатор группы, в которую необходимо добавить плейлист. (int, необязательно)
        :return: идентификатор плейлиста, если плейлист успешно добавлен, `None` в противном случае.
        """

        if not ownerId:
            if not groupId:
                return None

            ownerId = self.getSelf().get("id")

        response = self._VKReq("followPlaylist", {"owner_id": ownerId, "playlist_id": playlistId, **({"group_id": groupId} if groupId else {})})
        return response.get("id") if not (isinstance(response, dict) and response.get("error")) else response


    def removePlaylist(self, playlistId: int, groupId: int = None) -> Union[bool, dict]:
        """
        Удаляет плейлист из музыки пользователя или группы.

        Пример использования:\n
        result = client.removePlaylist(playlistId="yourPlaylistId", groupId="yourGroupId")\n
        print(result)

        :param playlistId: идентификатор плейлиста, который необходимо удалить. (int)
        :param groupId: идентификатор группы, из которой необходимо удалить плейлист. (int, необязательно)
        :return: `True`, если плейлист успешно удалён, `False` в противном случае.
        """

        if not groupId:
            groupId = self.getSelf().get("id")

        response = self._VKReq("deletePlaylist", {"owner_id": groupId, "playlist_id": playlistId})
        return bool(response) if not (isinstance(response, dict) and response.get("error")) else (response if response.get("error").get("code") != 17 else self._raiseError("playlistNotFound"))


    def _editPlaylistPhoto(self, playlistId: int, photo: str = None, groupId: int = None) -> Union[bool, dict]:
        if photo is None:
            return False

        if not groupId:
            groupId = self.getSelf().get("id")

        params = {"playlist_id": playlistId, "owner_id": groupId}
        if photo:
            method = "setPlaylistCoverPhoto"
            params = {**params, **{"photo": photo}}

        else:
            method = "deletePlaylistCoverPhoto"

        response = self._VKReq(method, params)
        return bool(response) if not (isinstance(response, dict) and response.get("error")) else response


    def editPlaylist(self, playlistId: int, title: Union[str, int] = None, description: Union[str, int] = None, photo: str = None, groupId: int = None) -> Union[bool, dict]:
        """
        Изменяет информацию плейлиста, принадлежащего пользователю или группе.

        Пример использования:\n
        result = client.editPlaylist(playlistId="yourPlaylistId", title="prombl — npc", description="Release Date: December 24, 2021", photo="yourPhotoUrl", groupId="yourGroupId")\n
        print(result)

        :param playlistId: идентификатор плейлиста, информацию которого необходимо изменить.
        :param title: новое название плейлиста. (Необязательно)
        :param description: новое описание плейлиста. (Необязательно)
        :param photo: новое фото плейлиста. (Необязательно)
        :param groupId: идентификатор группы, в которой находится плейлист. (Необязательно)
        :return: `True`, если информация плейлиста успешно обновлена, `False` в противном случае.
        """

        if not any((title, description is not None, photo is not None)):
            return False

        if not groupId:
            groupId = self.getSelf().get("id")

        params = {
            "playlist_id": playlistId,
            "owner_id": groupId,
            **({"title": title} if title else {}),
            **({"description": description} if description is not None else {})
        }

        if not any(("title" in params, "description" in params, photo is not None)):
            return False

        else:
            if any(("title" in params, "description" in params)):
                response = self._VKReq("editPlaylist", params)
                editInfoStatus = bool(response) if not (isinstance(response, dict) and response.get("error")) else response

            else:
                editInfoStatus = None

        if any((editInfoStatus is True, editInfoStatus is None)) and photo is not None:
            editPhotoStatus = self._editPlaylistPhoto(playlistId, photo, groupId)
            if editInfoStatus is None:
                editInfoStatus = editPhotoStatus

        return editInfoStatus


    def copyPlaylist(self, playlistId: int, ownerId: int = None, groupId: int = None, skipTitle: bool = False, skipDescription: bool = False, skipPhoto: bool = False) -> Union[int, None, dict]:
        """
        Копирует плейлист, принадлежий пользователю или группе в музыку пользователя или группы.

        Пример использования:\n
        result = client.copyPlaylist(playlistId=1, ownerId=-215973356, groupId: "yourGroupId", skipTitle=True, skipDescription=True, skipPhoto=True)\n
        print(result)

        :param playlistId: идентификатор плейлиста, который необходимо скопировать. (int)
        :param ownerId: идентификатор владельца плейлиста (пользователь или группа). (int, по умолчанию текущий пользователь)
        :param groupId: идентификатор группы, в которую необходимо скопировать плейлист. (int, необязательно)
        :param skipTitle: флаг, указывающий, необходимо ли не копировать название плейлиста (если `True`, будут использованы текущие дата и время). (bool, по умолчанию `False`)
        :param skipDescription: флаг, указывающий, необходимо ли не копировать описание плейлиста. (bool, по умолчанию `False`)
        :param skipPhoto: флаг, указывающий, необходимо ли не копировать фото плейлиста. (bool, по умолчанию `False`)
        :return: идентификатор плейлиста, если плейлист успешно скопирован, `None` в противном случае.
        """

        if not ownerId:
            ownerId = self.getSelf().get("id")

        playlist = self.getPlaylist(playlistId, ownerId, True)

        if isinstance(playlist, dict):
            if playlist.get("error"):
                return playlist

            title = playlist.get("title")
            description = playlist.get("description")
            photo = playlist.get("photo")

        else:
            title = playlist.title
            description = playlist.description if hasattr(playlist, "description") else None
            photo = playlist.photo if hasattr(playlist, "photo") else None

        if skipTitle:
            title = datetime.utcnow().replace(tzinfo=pytz.utc).astimezone(moscowTz).strftime("%d.%m.%Y / %H:%M:%S")

        if skipDescription:
            description = None

        if skipPhoto:
            photo = None

        elif photo:
            _, photo = photo.popitem()

        newPlaylistId = self.createPlaylist(title, description, photo, groupId)
        if isinstance(newPlaylistId, dict) and newPlaylistId.get("error"):
            return newPlaylistId

        tracks = (playlist.tracks if hasattr(playlist, "tracks") else None) if isinstance(playlist, Playlist) else playlist.get("tracks") if isinstance(playlist, dict) else None
        if not tracks:
            return newPlaylistId

        ownerIds = []
        trackIds = []
        for track in tracks[::-1]:
            ownerId, trackId = (track.ownerId, track.trackId) if isinstance(track, Track) else (track.get("ownerId"), track.get("trackId"))
            ownerIds.append(ownerId)
            trackIds.append(trackId)
        self.add(ownerIds, trackIds, newPlaylistId, groupId)

        return newPlaylistId


    def reorder(self, trackId: int, beforeTrackId: int = None, afterTrackId: int = None) -> Union[bool, dict]:
        """
        Изменяет порядок аудиотрека в музыке пользователя. Должен быть заполнен один из типов аргументов на выбор: `beforeTrackId` или `afterTrackId`.

        Пример использования для перемещения на место перед определённым треком:\n
        result = client.reorder(trackId="yourTrackId", beforeTrackId="yourBeforeTrackId")\n
        print(result)

        Пример использования для перемещения на место после определённого трека:\n
        result = client.reorder(trackId="yourTrackId", afterTrackId="yourAfterTrackId")\n
        print(result)

        :param trackId: идентификатор аудиотрека, порядок которого необходимо изменить. (int)
        :param beforeTrackId: идентификатор аудиотрека перед которым необходимо поместить аудиотрек. (int, необязательно)
        :param afterTrackId: идентификатор аудиотрека после которого необходимо поместить аудиотрек. (int, необязательно)
        :return: `True`, если порядок трека успешно изменён, `False` в противном случае.
        """

        if not any((beforeTrackId, afterTrackId)):
            return self._raiseError("trackReorderNeedsBeforeOrAfterArgument")

        if all((beforeTrackId, afterTrackId)):
            return self._raiseError("trackReorderNeedsOnlyBeforOrAfterNotBoth")

        response = self._VKReq("reorder", {"audio_id": trackId, **({"before": beforeTrackId} if beforeTrackId else {"after": afterTrackId})})
        return bool(response) if not (isinstance(response, dict) and response.get("error")) else response


    def followArtist(self, artistId: int = None) -> Union[bool, dict]:
        """
        Подписывается на обновления музыки артиста.

        Пример использования:\n
        result = client.followArtist(artistId=5696274288194638935)\n
        print(result)

        :param artistId: идентификатор артиста, на обновления которого необходимо подписаться. (int)
        :return: `True`, если Вы успешно подписались на обновления музыки артиста, `False` в противном случае.
        """

        response = self._VKReq("followArtist", {"artist_id": artistId})
        return bool(response) if not (isinstance(response, dict) and response.get("error")) else response


    def unfollowArtist(self, artistId: int = None) -> Union[bool, dict]:
        """
        Отписывается от обновлений музыки артиста.

        Пример использования:\n
        result = client.unfollowArtist(artistId=5696274288194638935)\n
        print(result)

        :param artistId: идентификатор артиста, от обновлений которого необходимо отписаться. (int)
        :return: `True`, если Вы успешно отписались от обновлений музыки артиста, `False` в противном случае.
        """

        response = self._VKReq("unfollowArtist", {"artist_id": artistId})
        return bool(response) if not (isinstance(response, dict) and response.get("error")) else response


    def followCurator(self, curatorId: int = None) -> Union[bool, dict]:
        """
        Подписывается на обновления музыки куратора.

        Пример использования:\n
        result = client.followCurator(curatorId=28905875)\n
        print(result)

        :param curatorId: идентификатор куратора, на обновления которого необходимо подписаться. (int)
        :return: `True`, если Вы успешно подписались на обновления музыки куратора, `False` в противном случае.
        """

        response = self._VKReq("followCurator", {"curator_id": curatorId})
        return bool(response) if not (isinstance(response, dict) and response.get("error")) else response


    def unfollowCurator(self, curatorId: int = None) -> Union[bool, dict]:
        """
        Отписывается от обновлений музыки куратора.

        Пример использования:\n
        result = client.unfollowCurator(curatorId=28905875)\n
        print(result)

        :param curatorId: идентификатор куратора, от обновлений которого необходимо отписаться. (int)
        :return: `True`, если Вы успешно отписались от обновлений музыки куратора, `False` в противном случае.
        """

        response = self._VKReq("unfollowCurator", {"curator_id": curatorId})
        return bool(response) if not (isinstance(response, dict) and response.get("error")) else response


    def followOwner(self, ownerId: int = None) -> Union[bool, dict]:
        """
        Подписывается на обновления музыки пользователя или группы.

        Пример использования:\n
        result = client.followOwner(ownerId=-215973356)\n
        print(result)

        :param ownerId: идентификатор пользователя или группы, на обновления которого(ой) необходимо подписаться. (int)
        :return: `True`, если Вы успешно подписались на обновления музыки пользователя или группы, `False` в противном случае.
        """

        response = self._VKReq("followOwner", {"owner_id": ownerId})
        return bool(response) if not (isinstance(response, dict) and response.get("error")) else response


    def unfollowOwner(self, ownerId: int = None) -> Union[bool, dict]:
        """
        Отписывается от обновлений музыки пользователя или группы.

        Пример использования:\n
        result = client.unfollowOwner(ownerId=-215973356)\n
        print(result)

        :param ownerId: идентификатор пользователя или группы, от обновлений которого(ой) необходимо отписаться. (int)
        :return: `True`, если Вы успешно отписались на обновлений музыки пользователя или группы, `False` в противном случае.
        """

        response = self._VKReq("followOwner", {"owner_id": ownerId})
        return bool(response) if not (isinstance(response, dict) and response.get("error")) else response


    def setBroadcast(self, ownerId: int = None, trackId: int = None, groupIds: Union[List[str], str] = None) -> Union[bool, dict]:
        """
        Устанавливает (удаляет) аудиотрек в (из) статус(а) пользователя или группы.

        Пример использования для установки аудиотрека в статус пользователя:\n
        result = client.setBroadcast(ownerId=474499156, trackId=456637846)\n
        print(result)

        Пример использования для установки аудиотрека в статус группы:\n
        result = client.setBroadcast(ownerId=474499156, trackId=456637846, groupdIds="yourGroupId")\n
        print(result)

        Пример использования для удаления аудиотрека из статуса пользователя:\n
        result = client.setBroadcast()\n
        print(result)

        Пример использования для удаления аудиотрека из статуса группы:\n
        result = client.setBroadcast(groupdIds="yourGroupId")\n
        print(result)

        :param ownerId: идентификатор владельца аудиотрека (пользователь или группа). (int, необязательно)
        :param trackId: идентификатор аудиотрека, который необходимо установить в статус. (int, необязательно)
        :param groupIds: идентификатор(ы) групп(ы), в (из) статус(а) которой необходимо установить (удалить) аудиотрек. (int, по умолчанию текущий пользователь)
        :return: `True`, если аудиотрек успешно установлен (удалён) в (из) статус(а), `False` в противном случае.
        """

        response = self._VKReq("setBroadcast", {**({"audio": f"{ownerId}_{trackId}"} if ownerId and trackId else {}), **({"target_ids": groupIds} if groupIds else {})})
        return True if not (isinstance(response, dict) and response.get("error")) else response